### X - 2017 commune

## question 1

def membre(x, q):
    """renvoie True si le point x est dans l'ensemble représenté par la liste q et 
False dans le cas contraire."""
    for i in range(len(q)):
        if q[i][0] == x[0] and q[i][1] == x[1]: return True
    return False

## question 2

def intersection(p, q):
    """renvoie une liste représentant l’intersection des
ensembles représentés par p et q."""
    inter = [] # va contenir l'intersection de p et q
    for i in range(len(p)):
        if membre(p[i], q):
            inter.append(p[i])
    return inter

## question 8

def bits(x, k):
    """renvoie la valeur du bit de coefficient 2^k dans la représentation binaire de
x."""
    b = bin(x)[2:]
    n = len(b)
    return int(b[n - k - 1]) if k < n else 0

def code(n, p):
    """prend en arguments
* un entier strictement positif n
* un point p représenté par une liste de longueur 2 dont les deux coordonnées sont prises
dans Dn.
Renvoie le codage de Lebesgue de p représenté sous la forme d’une liste Python."""
    x, y = p
    codage = []
    for k in range(n - 1, -1, -1):
        xk, yk = bits(x, k), bits(y, k)
        codage.append(xk * 2 + yk)
    return codage

## question 10

def compare_pcodes(n, c1, c2):
    """prend en arguments deux codages de Lebesgue de Dn × Dn.
Renvoie :
* 0 s’il sont égaux,
* 1 si c2 est plus grand par l'odre lexicographique que c1
* −1 sinon."""
    for i in range(n):
        # ici, c1[:i] = c2[:i]
        if c1[i] < c2[i]: return -1
        elif c2[i] < c1[i]: return 1
    # arrivé ici, c1 = c2
    return 0

## question 13

def ksuffixe(n, k, q):
    """prend en arguments
* un entier n strictement positif,
* une liste q représentant le codage de Lebesgue compacté d'un quadrant de Dn × Dn
* un entier naturel k inférieur strictement à n.

Si les k derniers chiffres de la liste q ont pour valeur 4, cette fonction renvoie
une nouvelle liste semblable à la liste q mais dont les k + 1 derniers chiffres
valent 4.
Sinon, cette fonction renvoie q inchangée."""
    new_q = list(q) # copie de q
    nb_4 = 0 # comptage des 4
    while nb_4 < k and q[n - 1 - nb_4] == 4: nb_4 += 1
    if nb_4 == k: new_q[n - 1 - nb_4] = 4
    return new_q

# Pour tester
print(ksuffixe(4, 2, [0, 1, 4, 4]))
print(ksuffixe(4, 2, [0, 1, 2, 4]))

## question 14

def compacte(n, s):
    """prend en arguments un entier strictement positif n et un ensemble de points P de
Dn x Dn représenté par la liste triée s des codages de Lebesgue de ses points, et
renvoie l'AQL de l'ensemble de points P ."""
    aql = s
    for k in range(n):
        aql_temp = [] # construction du k-eme compactage 
        m = len(aql)
        i = 0
        while i < m - 3:
            suffixe = ksuffixe(n, k, aql[i])
            suffixe1 = ksuffixe(n, k, aql[i + 3])
                    
            # Si les ksuffixes sont égaux, on a le quadrant complet 
            # (vu l'ordre sur les points).
            if compare_pcodes(n, suffixe, suffixe1) == 0:
                aql_temp.append(suffixe) # on compacte
                i += 4
            else: # pas de compactage
                aql_temp.append(aql[i])
                i += 1

        for j in range(i, m): # on ajoute les derniers non compactables
            aql_temp.append(aql[j])
            
        aql = aql_temp
        
    return aql

# Test
print(compacte(2, [[0, 0], [0, 3], [1, 2], [3, 0], [3, 1], [3, 2], [3, 3]]))

## Question 15

def compare_ccodes(n, p, q):
    compteur = 0 # combien d'éléments égaux au début ?
    while compteur < n and p[compteur] == q[compteur]:
        compteur += 1
        
    if compteur == n:      # quadrants égaux
        return 0
    
    elif q[compteur] == 4: # P inclus dans Q
        return 2

    elif p[compteur] == 4: # Q inclus dans P
        return -2
        
    elif p[compteur] < q[compteur]: # quadrants disjoints
        return 1           # p < q
    else:
        return -1          # q > p

print(compare_ccodes(3,[1,4,4],[2,4,4]))
print(compare_ccodes(3,[1,2,4],[1,4,4]))

## Question 16

def intersection2(n, p, q):
    """prend en argument un entier strictement positif n ainsi que deux AQL p et q 
représentant respectivement deux ensembles P et Q de points de Dn x Dn. 
Cette fonction renvoie un AQL représentant P inter Q."""
    inter = []
    np = len(p)
    nq = len(q)
    ip = 0              # indice de parcours de p
    iq = 0              # indice de parcours de q

    while ip < np and iq < nq:

        c = compare_ccodes(n, p[ip], q[iq]))

        if c == 0:      # quadrants égaux
            inter.append(p[ip])
            ip += 1
            iq += 1

        elif c == 2:    # p[ip] inclus dans q[iq]
            inter.append(p[ip])
            ip += 1
        elif c == -2:    # q[iq] inclus dans p[ip]
            inter.append(q[iq])
            iq += 1

        elif c == 1:    # quadrants disjoints, p[ip] < q[iq]
            ip += 1
        elif c == -1:   # quadrants disjoints, q[iq] < p[ip]
            iq += 1

    return inter
